//модуль для клиента
'use strict';
// это клиент

// Зависимости
let app = require('express')();
let timestamp = require('time-stamp');
const logSymbols = require('log-symbols');
let _ = require('lodash');
let ip = require('ip');
let nanoid = require('nanoid');
let axios = require('axios');
const bodyParser   = require('body-parser');

// Конфигурация
const wsName = require('../../package.json').name;
const wsVersion = require('../../package.json').version;
const wsClientConfig = require('../../wsClientConfig.json');
const wsCoreKey = wsClientConfig.wsCoreKey;
const wsPort = wsClientConfig.wsPort;
const wsEventBusAddress = wsClientConfig.wsEventBusAddress;
let apigwPath = null;
const instanceID = nanoid(10);

// Чтение body и cookie
app.use(bodyParser.json());
app.use(bodyParser.urlencoded({ extended: true}));
//app.use(cookieParser());

// Характеристики инстанса в сессии сокет-соединения
let socket = require('socket.io-client')(wsEventBusAddress,{
    query: {
        token: wsCoreKey, //jwt
        wsName: wsName,
        instanceID: instanceID
    }
});

let awaitedEvents = [];
let awaitedRequests = [];

let wsStartData = {
    wsName: wsName,
    wsVersion: wsVersion,
    awaitedEvents: awaitedEvents,
    awaitedRequests: awaitedRequests,
    instanceParams:{
        instanceID: instanceID,
        instanceIP: ip.address(),
        instancePort: wsPort}
};


// Отправляем данные о запущенном сервисе
// При успешном коннекте
socket.on('connect', function() {
    socket.emit('wsStart', wsStartData);
    console.log(wsStartData);
});

// При успешном коннекте
socket.on('afterCoreRestart', function(data) {
    //console.log(data);
    socket.emit('wsStart', wsStartData);
    console.log(wsStartData);
});

socket.on('coreIsDisabled', function(data) {
    console.log(data);
});


socket.on('regiseredInSystem', function(data) {
    console.log(data);
});

socket.on('error', (error) => {
    console.log(error)
});

socket.on('apigwPath', function(data) {
    apigwPath = data.apigwPath
});

// let on = function (eventName, ...temps) {
async function on (eventName, ...temps) {
    let right = null;
    let onEventAction = null; // функция передаваемая в обработчик выова
    if (typeof temps[0] === 'object') {
        console.log('Есть параметры, проверяем кому доступен/или кому не доступен');
        right = temps[0];
        onEventAction = temps[1];
    } else {
        console.log('Доступен всем');
        onEventAction = temps[0];
    }

    let personalEventName = `${wsStartData.instanceID}/${eventName}`;
    //console.log(personalEventName);
    function step1 (data){
        //console.log(`Получено сообщение с идентификатором: ${data.eventId}`);
        let dataToChangeEventState = {'eventId':data.eventId,'eventState':'received'};
        socket.emit('changeEventState', dataToChangeEventState);
    }
    async function step2 (data){
        try {
            await onEventAction(data);
            step3(data);
        } catch (err) {
            //console.log(`Ошибка обработки по сообщению с идентификатором: ${data.eventId}`);
            let dataToChangeEventState = {'eventId':data.eventId,'eventState':'error'};
            socket.emit('changeEventState', dataToChangeEventState);
        }
    }
    function step3 (data){
        //console.log(`Выполнена обработка по сообщению с идентификатором: ${data.eventId}`);
        let dataToChangeEventState = {'eventId':data.eventId,'eventState':'succsess'};
        socket.emit('changeEventState', dataToChangeEventState);
    }
    function run (data){
        step1(data);
        step2(data);
    }
    // Непосредственно сама подписка на событие 'instanceId\eventName' между шиной и сервисом
    socket.on(personalEventName, function(data){
        run(data)
    });
    // Регистрация подписки в MQ всех обработчиков, принадлежащих сервису
    // !!! todo Здесь добавление в awaitedEvents

    awaitedEvents.push({eventName:eventName, right: right}) // не забыть что был ещё personalEventName: personalEventName ()
}

function emit (eventName, ...temps) {
    // !!! todo Здесь заменяем на добавление в emittedEvents
    //console.log('Вызывается только в момент обращения к методу req');
    //todo добавить rights
    let right = null;
    let data = null;
    if (temps.length === 2) {
        //console.log('Есть параметры, проверяем кому доступен/или кому не доступен');
        right = temps[0];
        data = temps[1];
    } else {
        //console.log('Доступен всем');
        data = temps[0];
    }
    let eventData = {'eventName':eventName, right, data}
    //console.log(eventData);
    socket.emit('universalEvent', eventData);
}

// Проверка что запрос пришел именно через apigw
app.use(function(req, res, next){
    if (req.method ==='OPTIONS'){
        next();
    }
    else {
        //console.log(`Промежуточная функция до вызова метода`);
        if (req.method ==='GET' && req.query.instanceID === instanceID || (req.method ==='POST' && req.body.instanceID === instanceID)){
            next();
        } else {
            return res.send({result: 'error', reason: 'Запрос к сервису минуя apigw'})
        }

    }
});

async function get (name,...temps) {// availableTo = null, blockedTo = null, next
    //console.log({type: 'GET', path: name});
    let right = null;
    let next = null; // функция передаваемая в обработчик выова
    if (typeof temps[0] === 'object') {
        console.log('Есть параметры, проверяем кому доступен/или кому не доступен');
        right = temps[0];
        next = temps[1];
    } else {
        console.log('Доступен всем');
        next = temps[0];
    }
    // !!! todo Здесь заменяем на добавление в emittedEvents
    console.log(right);
    awaitedRequests.push({methodName:name, type: 'GET', right: right})
    //socket.emit('regRights', {wsName:wsStartData.wsName, rightsParams:rightsParams});
    app.get(name, next)
}
// добавить обработку правил
async function post (name, availableTo, next) {

    console.log({type: 'POST', path: name});
    //console.log(availableTo);
    app.post(name, next)
}

async function callPost (wsNameAndMethod, axiParams ) {
    let allParams = {fromWsName:wsName, fromInstanceID:instanceID, reqType:'ws'};
    if (axiParams !== undefined) {
        Object.assign(allParams, axiParams);
    }
    let fullPath = `${apigwPath}/${wsNameAndMethod}`;
    const axiosRes = await axios.post(fullPath, allParams);
    return axiosRes.data
}

async function callGet (wsNameAndMethod, axiParams ) {
    let allParams = {fromWsName:wsName, fromInstanceID:instanceID, reqType:'ws'};
    if (axiParams !== undefined) {
        Object.assign(allParams, axiParams);
    }
    let fullPath = `${apigwPath}/${wsNameAndMethod}`;
    const axiosRes = await axios.get(fullPath, {params:allParams});
    return axiosRes.data
}

async function callAxiGet (axiPath, axiParams ) {
    let allParams = {fromWsName:wsName, fromInstanceID:instanceID};
    if (axiParams !== undefined) {
        Object.assign(axiParams, allParams);
    }
    const axiosRes = await axios.get(axiPath, {params:allParams});
    return axiosRes.data
}

async function callAxiPost (axiosPath) {
    const axiosRes = await axios.post(axiosPath);
    return axiosRes.data
}


app.listen(wsPort, function(){
    console.log(`${timestamp('[DD-MM-YYYY HH:mm:ss.ms]')} ${logSymbols.success} Сервис ${wsName} запущен на ${wsPort} порту`)
});

module.exports = {
// Работа с запросами
    // Обработка входящих запросов
    get: get,
    post: post,
    // Для вызова межсервисных апи
    callGet: callGet,
    callPost: callPost,
    // Для вызова внешних апи
    callAxiGet: callAxiGet,
    callAxiPost: callAxiPost,
// Работа с событиями
    on: on,
    emit: emit
};

